/* mountmon - monitor mount/umount events
 * (C) James Budiono 2017, 2019
 * License: GPL v3
 */
#include <sys/epoll.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <poll.h>
#include <sys/types.h>
#include "mba/diff.h"
#include "mba/msgno.h"
#include "mba/varray.h"

#define MAX_BUF     (512*1024)  /* Typical mount in my system is 1KB. This is 512 times bigger */
#define MAX_EVENTS  1 /* Maximum number of events to be returned from a single epoll_wait() call */
#define PROC_MOUNTS "/proc/mounts"

void print_varray(struct varray *array) {
	iter_t iter;
	varray_iterate(array, &iter);

	char **line;
	while (line = varray_next(array, &iter)) {
		if (*line)
			printf("%s", *line);
	}
}

int varray_count(struct varray *array) {
	int count = 0;
	iter_t iter;
	varray_iterate(array, &iter);

	char **line;
	while (line = varray_next(array, &iter)) {
		if (*line)
			count++;
	}
	return count;
}

const void *linediff_idx(const void *s, int idx, void *context) {
	char **line = varray_get((struct varray*)s, idx);
	//printf("%s", *line);
	return *line;
}
int linediff_cmp(const void *e1, const void *e2, void *context) {
	//printf("%s", e1);
	//printf("%s", e2);	
	if ( (!e1) || (!e2) ) return 1;
	return strcmp(e1, e2);
}
  

void linediff(struct varray *old, struct varray *new) {
	int oldlen = varray_count(old);
	int newlen = varray_count(new);
	//printf("%d %d\n", oldlen, newlen);	
	//printf("OLD\n===\n");print_varray(old);printf("\n");
	//printf("NEW\n===\n");print_varray(new);printf("\n");
	
	struct varray *ses = varray_new(sizeof(struct diff_edit), NULL);
	int sn, d;
	
	if ((d = diff(old, 0, oldlen, new, 0, newlen, linediff_idx, linediff_cmp, NULL, 0, ses, &sn, NULL)) == -1) {
		MMNO(errno);
		return;
	}

	//printf("d=%d sn=%d\n", d, sn);
	int i;
	char **line;
	for (i = 0; i < sn; i++) {
		struct diff_edit *e = varray_get(ses, i);

		switch (e->op) {
/*					
			case DIFF_MATCH:
			//	printf("MAT: ");
			//	fwrite(old + e->off, 1, e->len, stdout);
				while (e->len) {
					line = varray_get(old, e->off);
					if (*line)
						printf("MAT: %s", *line);
					e->off++;
					e->len--;
				}				
				break;
*/		
			case DIFF_INSERT:
				//printf("INS: ");
				//fwrite(new + e->off, 1, e->len, stdout);				
				while (e->len) {
					line = varray_get(new, e->off);
					if (*line)
						printf("INS: %s", *line);
					e->off++;
					e->len--;
				}
				break;
			case DIFF_DELETE:
				//printf("DEL: ");
				//fwrite(old + e->off, 1, e->len, stdout);
				while (e->len) {
					line = varray_get(old, e->off);
					if (*line)
						printf("DEL: %s", *line);
					e->off++;
					e->len--;
				}				
				break;
		}
		//printf("\n");
	}
}

struct varray *split_to_lines(char *buf, int buflen) {	
	FILE *f = fmemopen(buf, buflen, "r");

	struct varray *ret = varray_new(sizeof(char *), NULL);
	int index = 0;	
	while (1) {
		size_t linelen = 0;		
		char **line = varray_get(ret, index);
		if (getline(line, &linelen, f) <= 0) {
			free(*line);
			*line = 0;
			break;
		}
		//printf("%d %s",linelen, *line);
		index++;
	}
	fclose(f);
	//print_varray(ret);
	
	return ret;
}

void delete_lines(struct varray *array) {
	iter_t iter;
	varray_iterate(array, &iter);

	char **line;
	while (line = varray_next(array, &iter)) {
		if (*line)
			free(*line);
	}
	varray_del(array);
}

int
main(int argc, char *argv[])
{
	struct varray *old_mount, *new_mount;
	char *buf = calloc (1, MAX_BUF);
	int buflen = 0;

	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);	
	
	// prime
	int fd = open(PROC_MOUNTS, O_RDONLY);
	if (fd == -1)
		perror("open");
	//fprintf(stderr,"Opened \"%s\" on fd %d\n", PROC_MOUNTS, fd);
	buflen = read(fd, buf, MAX_BUF-1);
	//fprintf(stderr,"read %d bytes: %.*s\n", buflen, buflen, buf);
	lseek(fd, 0, SEEK_SET);

	old_mount = split_to_lines(buf, buflen);

	// prepare for poll
	struct pollfd pfd;
	pfd.fd = fd;
	pfd.events = POLLERR | POLLPRI;

	// poll
	while (1) {
		poll(&pfd, 1, -1);
		
		buflen = read(fd, buf, MAX_BUF-1);
		if (buflen == -1)
			perror("read");
		//fprintf(stderr,"read %d bytes: %.*s\n", buflen, buflen, buf);
		lseek(fd, 0, SEEK_SET);			

		new_mount = split_to_lines(buf, buflen);
		linediff(old_mount, new_mount);
		
		delete_lines(old_mount);
		old_mount = new_mount;
	}
	
/*
	// prepare for epoll
    int epfd = epoll_create(1);
    if (epfd == -1)
		perror("epoll_create");

    struct epoll_event ev;
	ev.events = EPOLLPRI; // Only interested in read-priority events
	ev.data.fd = fd;
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, fd, &ev) == -1)
		perror("epoll_ctl");

	// epoll
    while (1) {
        //printf("About to epoll_wait()\n");
        		
		struct epoll_event evlist[MAX_EVENTS];
        int ready = epoll_wait(epfd, evlist, MAX_EVENTS, -1);
        if (ready == -1) {
            if (errno == EINTR)
                continue; //Restart if interrupted by signal
            else
                perror("epoll_wait");
        }
        //printf("Ready: %d\n", ready);

        //Deal with returned list of events
        int j;
        for (j = 0; j < ready; j++) {
            if (evlist[j].events & (EPOLLPRI)) {				
                buflen = read(evlist[j].data.fd, buf, MAX_BUF-1);
                if (buflen == -1)
                    perror("read");
                //printf("read %d bytes: %.*s\n", buflen, buflen, buf);
                lseek(evlist[j].data.fd, 0, SEEK_SET);                    

				new_mount = split_to_lines(buf, buflen);
                linediff(old_mount, new_mount);
                
				delete_lines(old_mount);
				old_mount = new_mount;
            } 
        }
    }
*/    
}

